﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace RatKing {

	public class Teacher : MonoBehaviour {
		public enum State {
			Walk,
			Idle,
			None
		}
		public enum StateTarget {
			WriteOnBlackboard,
			CreateATask,
			WalkAround,
			Watch,
			None
		}
		enum WPTarget {
			BlackboardOnly,
			NormalOnly,
			All
		}
		//
		public Transform waypointsTrans;
		public float minTargetDist = 0.1f;
		public float minTargetAngle = 10f;
		public float moveSpeed = 2f;
		public float rotSpeed = 180f;
		public float angrySeconds = 1f;
		public GameObject lightningPrefab;
		[Header("States")]
		public float waitTimeBlackboardWriting = 3f;
		public float waitTimeWalkingAroundMin = 5f;
		public float waitTimeWalkingAroundMax = 8f;
		[Header("Animation")]
		public Billboard billboard;
		public Anim[] anims;
		public GameObject particlesDust;
		public float walkTime = 0.5f;
		//
		List<Waypoint> waypointsNormal;
		List<Waypoint> waypointsBlackboard;
		State curState = State.None;
		StateTarget curStateTarget = StateTarget.None;
		Waypoint curWP;
		Waypoint[] curPath;
		int curPathIndex = -1;
		Vector3 curTargetPos;
		Vector3 lastPosition;
		Dictionary<string, Anim> animsByName = new Dictionary<string, Anim>();
		float waitTime;
		float angryTime;
		float walkTimer;

		//

		void Awake() {
			animsByName = new Dictionary<string, Anim>(2);
			for (int i = 0; i < anims.Length; ++i) {
				animsByName[anims[i].name] = anims[i];
			}
		}

		void Start() {
			waypointsNormal = new List<Waypoint>();
			waypointsBlackboard = new List<Waypoint>();
			foreach (var wp in waypointsTrans.GetComponentsInChildren<Waypoint>()) {
				if (wp.type == Waypoint.Type.Blackboard) {
					waypointsBlackboard.Add(wp);
				}
				else {
					waypointsNormal.Add(wp);
				}
			}
			curWP = Base.Helpers.DataStructure.GetRandomElement(waypointsNormal);
			if (curWP != null) {
				SetRandomPosition(curWP);
			}
			lastPosition = transform.position;
		}

		void Update() {
			billboard.velocity = (transform.position - lastPosition) * (1f / Time.deltaTime);
			lastPosition = transform.position;
			if (angryTime > 0f) {
				billboard.SetAnim(animsByName["angry"], 1f);
			}
			else if (billboard.velocity.sqrMagnitude > 0.01f) {
				billboard.SetAnim(animsByName["run"], 1f);

				if (walkTimer < Time.time) {
					if (particlesDust != null) {
						var p = transform.position;
						p.y = 0.5f;
						Instantiate(particlesDust, p, transform.rotation);
					}
					// Main.inst.sndWalk.Play(transform);
					walkTimer = Time.time + walkTime;
				}

			}
			else {
				billboard.velocity = new Vector3(curTargetPos.x - transform.position.x, 0f, curTargetPos.z - transform.position.z);
				billboard.SetAnim(animsByName["idle"]);
			}
			//
			if (angryTime > 0f) {
				angryTime -= Time.deltaTime;
				return;
			}
			if (waitTime > 0f) {
				waitTime -= Time.deltaTime;
				if (waitTime <= 0f) {
					FinishState();
				}
				// return;
			}
			switch (curState) {
				case State.Walk:
					UpdateWalk();
					break;
				case State.Idle:
					UpdateIdle();
					break;
			}
		}

		void UpdateWalk() {
			// check path point reached
			if (curPathIndex < curPath.Length && (curPathIndex == 0 || curPath[curPathIndex].Contains(transform.position, 0.1f))) {
				curPathIndex++;
				if (curPathIndex < curPath.Length) {
					// not reached target
					curWP = curPath[curPathIndex];
					curTargetPos = curWP.GetRandomPosition();
				}
			}
			var dir = new Vector3(curTargetPos.x - transform.position.x, 0f, curTargetPos.z - transform.position.z);
			var dist = dir.magnitude;
			// reached end?
			if (curPathIndex >= curPath.Length && dist < minTargetDist) {
				FinishState();
				return;
			}
			// rotate
			else { // if (dist >= minTargetDist) {
				dir.Normalize();
				var angle = Vector3.Angle(transform.forward, dir);
				float rot = rotSpeed * Time.deltaTime;
				transform.rotation = Quaternion.RotateTowards(transform.rotation, Quaternion.LookRotation(dir, Vector3.up), rot);
				// actually moving!
				if (angle < minTargetAngle) {
					var walkDist = Mathf.Min(Time.deltaTime * moveSpeed, dist);
					transform.Translate(Vector3.forward * walkDist, Space.Self);
				}
			}
		}

		void UpdateIdle() {
			if (curWP != null && (curWP.type == Waypoint.Type.Blackboard || curWP.type == Waypoint.Type.Window)) {
				// rotate to target
				curTargetPos = transform.position + Base.Helpers.Math.Rotate(Vector3.forward, Vector3.up, curWP.lookDirAngle) * 5f;
				var dir = new Vector3(curTargetPos.x - transform.position.x, 0f, curTargetPos.z - transform.position.z);
				dir.Normalize();
				float rot = rotSpeed * Time.deltaTime;
				transform.rotation = Quaternion.RotateTowards(transform.rotation, Quaternion.LookRotation(dir, Vector3.up), rot);
			}
		}

		//

		public bool GetAngryAt(Player player, int severity) {
			if (angryTime > 0f) {
				return false;
			}
			var seconds = angrySeconds * ((float)severity / 5f);
			angryTime = seconds;
			StartCoroutine(LightningCR(player, seconds));
			player.Freeze(seconds);
			player.AddScore(-severity);
			return true;
		}
		IEnumerator LightningCR(Player player, float seconds) {
			var l = (GameObject)Instantiate(lightningPrefab);
			float timer = seconds;
			while (timer > 0f) {
				timer -= Time.deltaTime;
				var p = transform.position + Vector3.up * 3f;
				Vector3 dir = (player.transform.position - p);
				float dist = dir.magnitude;
				dir.Normalize();
				l.transform.position = p;
				l.transform.rotation = Quaternion.LookRotation(dir, Base.Helpers.Math.Rotate(Vector3.up, dir, Random.Range(0f, 360f)));
				l.transform.localScale = new Vector3(1f, 1f, dist);
				foreach (var r in l.GetComponentsInChildren<Renderer>()) {
					r.material.mainTextureScale = new Vector2(1f, dist * 0.2f);
				}
				yield return null;
			}
			Destroy(l);
		}

		public bool IsLooking(Transform at = null) {
			if (at == null && Vector3.Angle(Vector3.back, transform.forward) < 110f) {
				return true;
			}
			else if (at != null && Vector3.Angle(transform.forward, (at.position - transform.position).normalized) < 110f) {
				return true;
			}
			return false;
		}

		public void GoToBlackboard() {
			if (FindPath(WPTarget.BlackboardOnly, true)) {
				SetState(State.Walk);
				curStateTarget = StateTarget.WriteOnBlackboard;
			}
		}

		public void SetState(State newState) {
			switch (newState) {
				case State.Idle:
					curPath = null;
					break;
				case State.Walk:
					break;
			}
			curState = newState;
		}

		void FinishState() {
			//print("state finished");
			switch (curStateTarget) {
				case StateTarget.WriteOnBlackboard:
					Main.inst.sndPotionWrite.Play(transform.position);
					//print("write on blackboard");
					SetState(State.Idle);
					waitTime = waitTimeBlackboardWriting;
					curStateTarget = StateTarget.CreateATask;
					break;
				case StateTarget.CreateATask: // wait for pupils to finish their tasks
					//print("wait for finished task");
					Main.inst.CreateTask();
					curStateTarget = StateTarget.WalkAround;
					waitTime = 0.001f;
					break;
				case StateTarget.Watch:
					//print("watch around");
					SetState(State.Idle);
					waitTime = Random.Range(waitTimeWalkingAroundMin, waitTimeWalkingAroundMin);
					curStateTarget = StateTarget.WalkAround;
					break;
				case StateTarget.WalkAround:
					//print("walk around");
					FindPath(WPTarget.NormalOnly, false);
					SetState(State.Walk);
					curStateTarget = StateTarget.Watch;
					break;
			}
		}

		//

		bool FindPath(WPTarget wpTarget, bool includeCur = false) {
			//print("find path");
			curPath = null;
			if (Waypoint.Pathfinding.Find(curWP, GetRandomWaypoint(wpTarget, includeCur), out curPath)) { // TEST
				curPathIndex = 0;
				curTargetPos = transform.position;
				return true;
			}
			return false;
		}

		void SetRandomPosition(Waypoint wp) {
			for(;;) {
				var pos = wp.GetRandomPosition();
				pos.y = 50f;
				RaycastHit hit;
				if (Physics.Raycast(pos, Vector3.down, out hit, 100f, (1 << LayerMask.NameToLayer("Floor")))) {
					transform.position = hit.point;
					return;
				}
			}
		}

		Waypoint GetRandomWaypoint(WPTarget target, bool includeCur) {
			var points = new List<Waypoint>();
			if (target == WPTarget.All || target == WPTarget.BlackboardOnly) {
				points.AddRange(waypointsBlackboard);
			}
			if (target == WPTarget.All || target == WPTarget.NormalOnly) {
				points.AddRange(waypointsNormal);
			}
			if (!includeCur) {
				points.Remove(curWP);
			}
			if (points.Count == 0) {
				Debug.LogError("Empty Waypoint List!");
				return null;
			}
			return Base.Helpers.DataStructure.GetRandomElement(points);
		}

		//

#if UNITY_EDITOR
		void OnDrawGizmos() {
			if (waypointsTrans != null) {
				Gizmos.color = new Color(1f, 1f, 0f, 0.5f);
				foreach (var wp in waypointsTrans.GetComponentsInChildren<Waypoint>()) {
					Gizmos.DrawLine(transform.position, wp.transform.position);
				}
			}
			if (curPath != null) {
				Gizmos.color = Color.yellow;
				for (int i = 1; i < curPath.Length; ++i) {
					Gizmos.DrawLine(curPath[i - 1].transform.position + Vector3.up, curPath[i].transform.position + Vector3.up);
				}
				if (curPathIndex >= 0) {
					Gizmos.color = Color.green;
					Gizmos.DrawLine(transform.position + transform.up, curTargetPos);
				}
			}
			Gizmos.color = Color.black;
			Gizmos.DrawSphere(curTargetPos, 0.3f);
		}
#endif
	}

}