﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

namespace RatKing {

	public class Waypoint : MonoBehaviour {
		public enum Type {
			Normal,
			Window,
			Blackboard
		}
		public Type type;
		public float lookDirAngle;
		public Waypoint[] neighbours;
		//
		BoxCollider coll;
		float[] dist;

		void OnValidate() {
			if (neighbours != null) {
				foreach (var wp in neighbours) {
					if (wp == this)
						Debug.LogWarning("assigned WP as its own neighbour!");
				}
			}
		}

		//

		void Awake() {
			coll = GetComponent<BoxCollider>();
			coll.enabled = false;
			dist = new float[neighbours.Length];
			for (int i = 0; i < neighbours.Length; ++i) {
				dist[i] = Vector3.Distance(transform.position, neighbours[i].transform.position);
			}
		}

		//

		public Vector3 GetRandomPosition() {
			var pos = Vector3.zero;
			pos.x += coll.size.x * Random.Range(-0.5f, 0.5f);
			pos.z += coll.size.z * Random.Range(-0.5f, 0.5f);
			return transform.TransformPoint(pos);
		}

		public bool Contains(Vector3 pos, float eps = 0f) {
			pos = transform.InverseTransformPoint(pos);
			return (pos.x >= -0.5f * coll.size.x - eps && pos.x <= 0.5f * coll.size.x + eps &&
				pos.z >= -0.5f * coll.size.z - eps && pos.z <= 0.5f * coll.size.z + eps);
		}

		//

#if UNITY_EDITOR
		void OnDrawGizmos() {
			Gizmos.color = type == Type.Blackboard ? Color.red : type == Type.Normal ? Color.green : Color.yellow;
			Gizmos.DrawSphere(transform.position, 0.3f);
			Gizmos.matrix = transform.localToWorldMatrix;
			Gizmos.DrawWireCube(Vector3.zero, Vector3.one);
			Gizmos.matrix = Matrix4x4.identity;
			Gizmos.color = Color.white;
			if (neighbours != null) {
				foreach (var wp in neighbours) {
					if (wp == null) continue;
					Gizmos.DrawLine(transform.position, wp.transform.position);
					var dir = (transform.position - wp.transform.position).normalized;
					Gizmos.DrawRay(wp.transform.position, 0.5f * Base.Helpers.Math.RotateAround(dir, Vector3.zero, wp.transform.up, 45f));
					Gizmos.DrawRay(wp.transform.position, 0.5f * Base.Helpers.Math.RotateAround(dir, Vector3.zero, wp.transform.up, -45f));
				}
			}
			if (type == Type.Blackboard || type == Type.Window) {
				Gizmos.color = Color.black;
				var bbDir = Base.Helpers.Math.RotateAround(transform.forward, Vector3.zero, transform.up, lookDirAngle);
				var coll = GetComponent<BoxCollider>();
				var l = Mathf.Max(transform.localScale.x * coll.size.x, transform.localScale.z * coll.size.z) * 0.55f;
				Gizmos.DrawRay(transform.position, l * bbDir);
				Gizmos.DrawRay(transform.position + l * bbDir, 0.35f * Base.Helpers.Math.RotateAround(-bbDir, Vector3.zero, transform.up, 45f));
				Gizmos.DrawRay(transform.position + l * bbDir, 0.35f * Base.Helpers.Math.RotateAround(-bbDir, Vector3.zero, transform.up, -45f));
			}
		}
#endif

		//

		public static class Pathfinding {
			class Node {
				public Node parent = null;
				public Waypoint wp;
				public float F = 0;
				public float G = 0;
				public float H = 0;
				public Node(Waypoint block) { this.wp = block; }
				public Node(Waypoint block, Node parent, float F, float G, float H) { this.wp = block; this.parent = parent; this.F = F; this.G = G; this.H = H; }
			};

			static float GetManhattanDistance(Vector3 a, Vector3 b) {
				return Mathf.Abs(a.x - b.x) + Mathf.Abs(a.y - b.y) + Mathf.Abs(a.z - b.z);
			}

			// Finds A Path Between Two Blocks, Returns The Path In An Array
			public static bool Find(Waypoint A, Waypoint B, out Waypoint[] path, int maxLength = 1000) {
				if (A == null || B == null) {
					path = null;
					return false;
				}

				path = null;
				var open = new List<Node>(maxLength);
				var closed = new List<Node>(maxLength);
				Node current = null;
				open.Add(new Node(A));

				// pathfinding!
				for (;;) {

					// no nodes any more, but target not found? -> end
					if (open.Count == 0)
						return false;

					current = open[0]; // list is sorted!

					// remove node from open list and move it into closed list
					open.Remove(current);
					closed.Add(current);

					// target found? -> end the search!
					if (current.wp == B) {
						int pathLength = 0;
						Node counter = current;

						do { // go backwards, for counting only
							counter = counter.parent;
							++pathLength;
						} while (counter != null);

						pathLength++;
						path = new Waypoint[pathLength];

						// go forward, for building path
						for (int i = 0; i < pathLength - 1; ++i) {
							path[pathLength - i - 1] = current.wp;
							current = current.parent;
						}
						path[0] = A;

						return true;
					}

					float G = current.G + 1; // Position.GetSqrDistance(current.block.absPosition, neighbour.absPosition);
					if (G > maxLength) continue; // return false;

					// iterate through connections
					int nm = current.wp.neighbours.Length;
					for (int i = 0; i < nm; ++i) {
						Waypoint neighBlock = current.wp.neighbours[i];
						if (neighBlock == null) // || (!ignoreCreatures && neighBlock != B && neighBlock.creature != null))
							continue;

						// node is already in closed list?
						if (closed.Find(n => n.wp == neighBlock) != null)
							continue;

						// node is already in open list?
						Node oldNode = open.Find(n => n.wp == neighBlock);
						if (oldNode != null) {
							if (G < oldNode.G) {
								oldNode.G = G;
								oldNode.F = G + oldNode.H;
								oldNode.parent = current;
								open.Sort((Node a, Node b) => Mathf.FloorToInt(a.F - b.F));
							}
						}
						else {
							float H = GetManhattanDistance(neighBlock.transform.position, B.transform.position);
							float F = G + H;
							Node newNode = new Node(neighBlock, current, F, G, H);
							open.Add(newNode);
							open.Sort((Node a, Node b) => Mathf.FloorToInt(a.F - b.F));
						}
					}
				}
			}
		}
	}

}